"""
 Copyright (c) 2019 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
import argparse
import psutil
import signal
import subprocess
import sys
import threading
import time
from python_banyan.banyan_base import BanyanBase

# import webbrowser


class S3P(BanyanBase, threading.Thread):
    """
    This class starts the Banyan server to support the Scratch 3 OneGPIO
    for the Picoboard

    It will start the backplane, picoboard gateway and websocket gateway.
    """

    def __init__(self, com_port=None, log=None):
        """

        :param com_port: Manually select serial com port
        :param log: log unhandled exceptions.
        """

        self.com_port = com_port
        self.log = log

        self.backplane_exists = False
        self.proc_bp = None
        self.proc_awg = None
        self.proc_hwg = None


        # start backplane
        self.start_backplane()

        # start the websocket gateway
        self.start_wsgw()

        # start picoboard gateway
        self.start_pbgw()


        # webbrowser.open('https://mryslab.github.io/s3onegpio/', new=1)

        # start the background scheduler to periodically run check_processes and confirm methods
        # self.scheduler = BackgroundScheduler()
        # self.job = self.scheduler.add_job(self.check_local_processes, 'interval', seconds=.1)
        # start the schedule running
        threading.Thread.__init__(self)
        self.daemon = True
        self.stop_event = threading.Event()
        self.stop_event.clear()
        self.start()

        while True:
            if self.stop_event.is_set():
                print('Exiting. Error detected - is your Picoboard plugged in?')
                sys.exit(0)
            try:
                time.sleep(.01)

            except KeyboardInterrupt:
                # self.proc_awg.kill()
                # self.proc_hwg.kill()
                # sys.exit(0)
                self.killall()

    def run(self):

        while not self.stop_event.is_set():
            z = [p.info for p in psutil.process_iter(attrs=['pid', 'name', 'status'])
                 if 'backplane' in p.info['name']]

            if not z:
                self.start_backplane()

            if z[0]['status'] == 'zombie':
                # self.start_backplane()
                self.stop_event.set()
                self.killall()
                # sys.exit(0)


            z = [p.info for p in psutil.process_iter(attrs=['pid', 'name', 'status'])
                 if 'wsgw' in p.info['name']]

            if not z:
                # self.start_wsgw()
                self.stop_event.set()
                self.killall()

            if z[0]['status'] == 'zombie':
                # self.start_wsgw()
                self.stop_event.set()
                self.killall()
            z = [p.info for p in psutil.process_iter(attrs=['pid', 'name', 'status'])
                 if 'pbgw' in p.info['name']]

            if not z:
                # self.start_pbgw()
                self.stop_event.set()
                self.killall()

            if z[0]['status'] == 'zombie':
                # print('pz')
                # time.sleep(4)
                # self.start_pbgw()
                self.stop_event.set()
                self.killall()
            time.sleep(.1)

    def killall(self):
        p = psutil.Process(self.proc_awg)
        p.kill()
        p = psutil.Process(self.proc_hwg)
        p.kill()
        p = psutil.Process(self.proc_bp)
        p.kill()
        sys.exit(0)


    def start_backplane(self):
        for pid in psutil.pids():
            p = psutil.Process(pid)
            if p.name() == "backplane":
                print("Backplane already started.          PID = " + str(p.pid))
                self.backplane_exists = True
                self.proc_bp = p.pid
            else:
                continue

        # if backplane is not already running, start a new instance
        if not self.backplane_exists:
            if sys.platform.startswith('win32'):
                self.proc_bp = subprocess.Popen('backplane',
                                                 creationflags=subprocess.CREATE_NEW_CONSOLE).pid

            else:
                self.proc_bp = subprocess.Popen('backplane',
                                                 stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                                                 stdout=subprocess.PIPE).pid
            self.backplane_exists = True
            print(self.proc_bp)

            print('Backplane is started...')

    def start_wsgw(self):
        if sys.platform.startswith('win32'):
            wsgw_start = ['wsgw', '-i', '9004']
        else:
            wsgw_start = ['wsgw', '-i', '9004']

        if self.log:
            wsgw_start.append('-l')
            wsgw_start.append('True')

        if sys.platform.startswith('win32'):
            self.proc_awg = subprocess.Popen(wsgw_start,
                                             creationflags=subprocess.CREATE_NEW_CONSOLE).pid

        else:
            self.proc_awg = subprocess.Popen(wsgw_start,
                                             stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                                             stdout=subprocess.PIPE).pid

        print(self.proc_awg)

        print('WebSocket Gateway is started...')


    def start_pbgw(self):
        if sys.platform.startswith('win32'):
            self.pbgw_start = ['pbgw']
        else:
            self.pbgw_start = ['pbgw']

        if self.log:
            self.pbgw_start.append('-l')
            self.pbgw_start.append('True')
        if self.com_port:
            self.pbgw_start.append('-c')
            self.pbgw_start.append(self.com_port)

        if sys.platform.startswith('win32'):
            self.proc_hwg = subprocess.Popen(self.pbgw_start,
                                             creationflags=subprocess.CREATE_NEW_CONSOLE).pid
        else:
            self.proc_hwg = subprocess.Popen(self.pbgw_start,
                                             stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                                             stdout=subprocess.PIPE).pid

        print(self.proc_hwg)


        print('Picoboard Gateway is started...')


def signal_handler(sig, frame):
    print('Exiting Through Signal Handler')
    raise KeyboardInterrupt


def s3px():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", dest="com_port", default="None",
                        help="Use this COM port instead of auto discovery")
    parser.add_argument("-l", dest="log", default="False",
                        help="Set to True to turn logging on.")

    args = parser.parse_args()

    log = args.log.lower()
    if log == 'false':
        log = False
    else:
        log = True

    if args.com_port == "None":
        com_port = None
    else:
        com_port = args.com_port

    S3P(com_port=com_port, log=log)


# listen for SIGINT
signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)

if __name__ == '__main__':
    # replace with name of function you defined above
    s3px()
